# Generated by Django 4.2.18 on 2025-02-04 12:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("projectroles", "0001_squashed_0032_alter_appsetting_value"),
    ]

    operations = [
        migrations.AlterField(
            model_name="sodaruseradditionalemail",
            name="secret",
            field=models.CharField(
                help_text="Secret token for email verification",
                max_length=255,
                null=True,
                unique=True,
            ),
        ),
    ]



# Generated by Django 4.2.14 on 2024-07-15 14:46

from django.conf import settings
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import markupfield.fields
import projectroles.models
import random
import string
import uuid


def populate_roles(apps, schema_editor):
    """Populate Role objects"""
    Role = apps.get_model('projectroles', 'Role')

    def save_role(name, description):
        role = Role(name=name, description=description)
        role.save()

    # Owner
    save_role(
        'project owner',
        'The project owner has full access to project data; rights to add, '
        'modify and remove project members; right to assign a project '
        'delegate. Each project must have exactly one owner.',
    )
    # Delegate
    save_role(
        'project delegate',
        'The project delegate has all the rights of a project owner with the '
        'exception of assigning a delegate. A maximum of one delegate can be '
        'set per project. A delegate role can be set by project owner.',
    )
    # Contributor
    save_role(
        'project contributor',
        'A project member with ability to view and add project data. Can edit '
        'their own data.',
    )
    # Guest
    save_role(
        'project guest',
        'Read-only access to a project. Can view data in project, can not add '
        'or edit.',
    )


def fix_remote_project_keys(apps, schema_editor):
    """Fix missing RemoteProject.project foreign keys (issue #197)"""
    RemoteProject = apps.get_model('projectroles', 'RemoteProject')
    Project = apps.get_model('projectroles', 'Project')
    for rp in RemoteProject.objects.all():
        if not rp.project:
            rp.project = Project.objects.filter(
                sodar_uuid=rp.project_uuid
            ).first()
            rp.save()


def populate_project_full_title(apps, schema_editor):
    """Populate the new full_title field in the Project model"""

    def get_project_parents(obj):
        if not obj.parent:
            return None
        ret = []
        parent = obj.parent
        while parent:
            ret.append(parent)
            parent = parent.parent
        return reversed(ret)

    def get_project_full_title(obj):
        parents = get_project_parents(obj)
        ret = ' / '.join([p.title for p in parents]) + ' / ' if parents else ''
        ret += obj.title
        return ret

    Project = apps.get_model('projectroles', 'Project')
    for project in Project.objects.all():
        project.full_title = get_project_full_title(project)
        project.save()


def populate_public_children(apps, schema_editor):
    """Populate the new has_public_children field in the Project model"""

    def get_public_children(obj):
        for child in obj.children.all():
            if child.public_guest_access:
                return True
            ret = get_public_children(child)
            if ret:
                return True
        return False

    Project = apps.get_model('projectroles', 'Project')
    for project in Project.objects.all():
        project.has_public_children = get_public_children(project)
        project.save()


def set_role_ranks(apps, schema_editor):
    """Set rank values for existing roles"""
    role_ranking = {
        'project owner': 10,
        'project delegate': 20,
        'project contributor': 30,
        'project guest': 40,
    }
    Role = apps.get_model('projectroles', 'Role')
    for role in Role.objects.all():
        if role.name in role_ranking:
            role.rank = role_ranking[role.name]
            role.save()


def migrate_project_stars(apps, schema_editor):
    """Create project_star AppSettings from ProjectUserTag objects"""
    ProjectUserTag = apps.get_model('projectroles', 'ProjectUserTag')
    AppSetting = apps.get_model('projectroles', 'AppSetting')
    for tag in ProjectUserTag.objects.filter(name='STARRED'):
        AppSetting.objects.get_or_create(
            project=tag.project,
            user=tag.user,
            value=True,
            name='project_star',
        )


def delete_category_app_settings(apps, schema_editor):
    """Delete app settings assigned to categories"""
    from projectroles.app_settings import AppSettingAPI

    app_settings = AppSettingAPI()
    AppSetting = apps.get_model('projectroles', 'AppSetting')
    # Find all app settings whith a project
    pr_settings = AppSetting.objects.exclude(project=None)
    for app_setting in pr_settings:
        try:
            plugin_name = (
                app_setting.app_plugin.name
                if app_setting.app_plugin
                else 'projectroles'
            )
            setting_def = app_settings.get_definition(
                plugin_name=plugin_name, name=app_setting.name
            )
        except ValueError:
            app_setting.delete()
            continue
        if app_setting.project.type not in setting_def.project_types:
            # Delete app setting if it is not restricted to any project types
            app_setting.delete()


def populate_finder_role(apps, schema_editor):
    """Populate project finder role"""
    Role = apps.get_model('projectroles', 'Role')
    role = Role(
        name='project finder',
        rank=50,
        project_types=['CATEGORY'],
        description='The project finder is able to see certain details of all '
        'categories and projects under a category for this role is given. '
        'They will not have access to the apps or data of those subprojects or '
        'categories until assigned a higher role. This role can only be '
        'assigned for categories.',
    )
    role.save()


def reverse_finder_role(apps, schema_editor):
    """Reverse code for project finder role populating"""
    Role = apps.get_model('projectroles', 'Role')
    role = Role.objects.filter(name='project finder').first()
    if role:
        role.delete()


def populate_additional_email_model(apps, schema_editor):
    """
    Move existing additional email app settings into the
    SODARUserAdditionalEmail model as verified emais. Delete the settings
    objects.
    """
    AppSetting = apps.get_model('projectroles', 'AppSetting')
    SODARUserAdditionalEmail = apps.get_model(
        'projectroles', 'SODARUserAdditionalEmail'
    )
    for a in AppSetting.objects.filter(
        app_plugin=None, name='user_email_additional'
    ):
        for v in a.value.split(';'):
            secret = ''.join(
                random.SystemRandom().choice(
                    string.ascii_lowercase + string.digits
                )
                for _ in range(32)
            )
            try:
                SODARUserAdditionalEmail.objects.create(
                    user=a.user, email=v, verified=True, secret=secret
                )
            except Exception:
                pass
        a.delete()


class Migration(migrations.Migration):

    replaces = [
        ('projectroles', '0001_initial'),
        ('projectroles', '0002_auto_20180411_1758'),
        ('projectroles', '0003_populate_roles'),
        ('projectroles', '0004_rename_uuid'),
        ('projectroles', '0005_update_uuid'),
        ('projectroles', '0006_add_remote_projects'),
        ('projectroles', '0007_fix_remoteproject_foreign_key'),
        ('projectroles', '0008_auto_20190426_0606'),
        ('projectroles', '0009_rename_projectsetting'),
        ('projectroles', '0010_update_appsetting'),
        ('projectroles', '0011_remove_indexes'),
        ('projectroles', '0012_update_remotesite'),
        ('projectroles', '0013_update_appsetting_type'),
        ('projectroles', '0014_update_appsetting_value_json'),
        ('projectroles', '0015_fix_appsetting_constraint'),
        ('projectroles', '0016_app_plugin_field_none'),
        ('projectroles', '0017_project_full_title'),
        ('projectroles', '0018_update_jsonfield'),
        ('projectroles', '0019_project_public_guest_access'),
        ('projectroles', '0020_project_has_public_children'),
        ('projectroles', '0021_remove_project_submit_status'),
        ('projectroles', '0022_role_rank'),
        ('projectroles', '0023_project_archive'),
        ('projectroles', '0024_project_star_app_setting'),
        ('projectroles', '0025_delete_projectusertag'),
        ('projectroles', '0026_delete_category_settings'),
        ('projectroles', '0027_role_project_type'),
        ('projectroles', '0028_populate_finder_role'),
        ('projectroles', '0029_sodaruseradditionalemail'),
        ('projectroles', '0030_populate_sodaruseradditionalemail'),
        ('projectroles', '0031_remotesite_owner_modifiable'),
        ('projectroles', '0032_alter_appsetting_value'),
    ]

    initial = True

    dependencies = [
        ('djangoplugins', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Project',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'title',
                    models.CharField(help_text='Project title', max_length=255),
                ),
                (
                    'type',
                    models.CharField(
                        choices=[
                            ('CATEGORY', 'Category'),
                            ('PROJECT', 'Project'),
                        ],
                        default='PROJECT',
                        help_text='Type of project ("CATEGORY", "PROJECT")',
                        max_length=64,
                    ),
                ),
                (
                    'description',
                    models.CharField(
                        help_text='Short project description', max_length=512
                    ),
                ),
                (
                    'readme',
                    markupfield.fields.MarkupField(
                        blank=True,
                        help_text='Project README (optional, supports markdown)',
                        null=True,
                        rendered_field=True,
                    ),
                ),
                (
                    'readme_markup_type',
                    models.CharField(
                        choices=[
                            ('', '--'),
                            ('html', 'HTML'),
                            ('plain', 'Plain'),
                            ('markdown', 'Markdown'),
                            ('restructuredtext', 'Restructured Text'),
                        ],
                        default='markdown',
                        editable=False,
                        max_length=30,
                    ),
                ),
                (
                    'submit_status',
                    models.CharField(
                        default='OK',
                        help_text='Status of project creation',
                        max_length=64,
                    ),
                ),
                (
                    '_readme_rendered',
                    models.TextField(editable=False, null=True),
                ),
                (
                    'omics_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='Project Omics UUID',
                        unique=True,
                    ),
                ),
                (
                    'parent',
                    models.ForeignKey(
                        blank=True,
                        help_text='Parent category if nested',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='children',
                        to='projectroles.project',
                    ),
                ),
            ],
            options={
                'ordering': ['parent__title', 'title'],
                'unique_together': {('title', 'parent')},
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='Name of role', max_length=64, unique=True
                    ),
                ),
                ('description', models.TextField(help_text='Role description')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectUserTag',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        default='STARRED',
                        help_text='Name of tag to be assigned',
                        max_length=64,
                    ),
                ),
                (
                    'omics_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='ProjectUserTag Omics UUID',
                        unique=True,
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project in which the tag is assigned',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='tags',
                        to='projectroles.project',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='User for whom the tag is assigned',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='project_tags',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': ['project__title', 'user__username', 'name'],
            },
        ),
        migrations.CreateModel(
            name='ProjectInvite',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'email',
                    models.EmailField(
                        help_text='Email address of the person to be invited',
                        max_length=254,
                    ),
                ),
                (
                    'date_created',
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text='DateTime of invite creation',
                    ),
                ),
                (
                    'date_expire',
                    models.DateTimeField(
                        help_text='Expiration of invite as DateTime'
                    ),
                ),
                (
                    'message',
                    models.TextField(
                        blank=True,
                        help_text='Message to be included in the invite email (optional)',
                    ),
                ),
                (
                    'secret',
                    models.CharField(
                        help_text='Secret token provided to user with the invite',
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    'active',
                    models.BooleanField(
                        default=True,
                        help_text='Status of the invite (False if claimed or revoked)',
                    ),
                ),
                (
                    'omics_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='ProjectInvite Omics UUID',
                        unique=True,
                    ),
                ),
                (
                    'issuer',
                    models.ForeignKey(
                        help_text='User who issued the invite',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='issued_invites',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project to which the person is invited',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='invites',
                        to='projectroles.project',
                    ),
                ),
                (
                    'role',
                    models.ForeignKey(
                        help_text='Role assigned to the person',
                        on_delete=django.db.models.deletion.CASCADE,
                        to='projectroles.role',
                    ),
                ),
            ],
            options={
                'ordering': ['project__title', 'email', 'role__name'],
            },
        ),
        migrations.CreateModel(
            name='RoleAssignment',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'omics_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='RoleAssignment Omics UUID',
                        unique=True,
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project in which role is assigned',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='roles',
                        to='projectroles.project',
                    ),
                ),
                (
                    'role',
                    models.ForeignKey(
                        help_text='Role to be assigned',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='assignments',
                        to='projectroles.role',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='User for whom role is assigned',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='roles',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': [
                    'project__parent__title',
                    'project__title',
                    'role__name',
                    'user__username',
                ],
                'indexes': [
                    models.Index(
                        fields=['project'],
                        name='projectrole_project_b13795_idx',
                    ),
                    models.Index(
                        fields=['user'], name='projectrole_user_id_0e46f7_idx'
                    ),
                ],
            },
        ),
        migrations.CreateModel(
            name='ProjectSetting',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='Name of the setting', max_length=255
                    ),
                ),
                (
                    'type',
                    models.CharField(
                        choices=[
                            ('BOOLEAN', 'Boolean'),
                            ('INTEGER', 'Integer'),
                            ('STRING', 'String'),
                        ],
                        help_text='Type of the setting',
                        max_length=64,
                    ),
                ),
                (
                    'value',
                    models.CharField(
                        blank=True,
                        help_text='Value of the setting',
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    'omics_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='ProjectSetting Omics UUID',
                        unique=True,
                    ),
                ),
                (
                    'app_plugin',
                    models.ForeignKey(
                        help_text='App to which the setting belongs',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='settings',
                        to='djangoplugins.plugin',
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project to which the setting belongs',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='settings',
                        to='projectroles.project',
                    ),
                ),
            ],
            options={
                'ordering': ['project__title', 'app_plugin__name', 'name'],
                'unique_together': {('project', 'app_plugin', 'name')},
            },
        ),
        migrations.RunPython(
            code=populate_roles,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RenameField(
            model_name='project',
            old_name='omics_uuid',
            new_name='sodar_uuid',
        ),
        migrations.RenameField(
            model_name='projectinvite',
            old_name='omics_uuid',
            new_name='sodar_uuid',
        ),
        migrations.RenameField(
            model_name='projectsetting',
            old_name='omics_uuid',
            new_name='sodar_uuid',
        ),
        migrations.RenameField(
            model_name='projectusertag',
            old_name='omics_uuid',
            new_name='sodar_uuid',
        ),
        migrations.RenameField(
            model_name='roleassignment',
            old_name='omics_uuid',
            new_name='sodar_uuid',
        ),
        migrations.AlterField(
            model_name='project',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4, help_text='Project SODAR UUID', unique=True
            ),
        ),
        migrations.AlterField(
            model_name='projectinvite',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text='ProjectInvite SODAR UUID',
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name='projectsetting',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text='ProjectSetting SODAR UUID',
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name='projectusertag',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text='ProjectUserTag SODAR UUID',
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name='roleassignment',
            name='sodar_uuid',
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text='RoleAssignment SODAR UUID',
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name='project',
            name='description',
            field=models.CharField(
                blank=True,
                help_text='Short project description',
                max_length=512,
                null=True,
            ),
        ),
        migrations.CreateModel(
            name='RemoteSite',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='Site name', max_length=255, unique=True
                    ),
                ),
                ('url', models.URLField(help_text='Site URL', max_length=2000)),
                (
                    'mode',
                    models.CharField(
                        default='TARGET', help_text='Site mode', max_length=64
                    ),
                ),
                ('description', models.TextField(help_text='Site description')),
                (
                    'secret',
                    models.CharField(
                        help_text='Secret token for connecting to the source site',
                        max_length=255,
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='RemoteSite relation UUID (local)',
                        unique=True,
                    ),
                ),
            ],
            options={
                'ordering': ['name'],
                'unique_together': {('url', 'mode', 'secret')},
            },
        ),
        migrations.CreateModel(
            name='RemoteProject',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'project_uuid',
                    models.UUIDField(default=None, help_text='Project UUID'),
                ),
                (
                    'level',
                    models.CharField(
                        default='NONE',
                        help_text='Project access level',
                        max_length=255,
                    ),
                ),
                (
                    'date_access',
                    models.DateTimeField(
                        help_text='DateTime of last access from/to remote site',
                        null=True,
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='RemoteProject relation UUID (local)',
                        unique=True,
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        blank=True,
                        help_text='Related project object (if created locally)',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='remotes',
                        to='projectroles.project',
                    ),
                ),
                (
                    'site',
                    models.ForeignKey(
                        help_text='Remote SODAR site',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='projects',
                        to='projectroles.remotesite',
                    ),
                ),
            ],
            options={
                'ordering': ['site__name', 'project_uuid'],
            },
        ),
        migrations.RunPython(
            code=fix_remote_project_keys,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AddField(
            model_name='projectsetting',
            name='user',
            field=models.ForeignKey(
                blank=True,
                help_text='User to which the setting belongs',
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='user_settings',
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name='projectsetting',
            name='project',
            field=models.ForeignKey(
                blank=True,
                help_text='Project to which the setting belongs',
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='settings',
                to='projectroles.project',
            ),
        ),
        migrations.CreateModel(
            name='AppSetting',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='Name of the setting', max_length=255
                    ),
                ),
                (
                    'type',
                    models.CharField(
                        choices=[
                            ('BOOLEAN', 'Boolean'),
                            ('INTEGER', 'Integer'),
                            ('STRING', 'String'),
                        ],
                        help_text='Type of the setting',
                        max_length=64,
                    ),
                ),
                (
                    'value',
                    models.CharField(
                        blank=True,
                        help_text='Value of the setting',
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='AppSetting SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'app_plugin',
                    models.ForeignKey(
                        help_text='App to which the setting belongs',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='settings',
                        to='djangoplugins.plugin',
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        blank=True,
                        help_text='Project to which the setting belongs',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='settings',
                        to='projectroles.project',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        blank=True,
                        help_text='User to which the setting belongs',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='user_settings',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'user_modifiable',
                    models.BooleanField(
                        default=True, help_text='Setting visibility in forms'
                    ),
                ),
                (
                    'value_json',
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict,
                        help_text='Optional JSON value for the setting',
                    ),
                ),
            ],
            options={
                'ordering': ['project__title', 'app_plugin__name', 'name'],
                # 'unique_together': {('project', 'app_plugin', 'name')},
            },
        ),
        migrations.DeleteModel(
            name='ProjectSetting',
        ),
        migrations.RemoveIndex(
            model_name='roleassignment',
            name='projectrole_project_b13795_idx',
        ),
        migrations.RemoveIndex(
            model_name='roleassignment',
            name='projectrole_user_id_0e46f7_idx',
        ),
        migrations.AddField(
            model_name='remotesite',
            name='user_display',
            field=models.BooleanField(
                default=True, help_text='RemoteSite visibility to users'
            ),
        ),
        migrations.AlterField(
            model_name='remotesite',
            name='secret',
            field=models.CharField(
                help_text='Secret token for connecting to the source site',
                max_length=255,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='appsetting',
            name='type',
            field=models.CharField(
                help_text='Type of the setting', max_length=64
            ),
        ),
        migrations.AlterField(
            model_name='appsetting',
            name='value_json',
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict,
                help_text='Optional JSON value for the setting',
                null=True,
            ),
        ),
        migrations.AlterUniqueTogether(
            name='appsetting',
            unique_together={('project', 'user', 'app_plugin', 'name')},
        ),
        migrations.AlterField(
            model_name='appsetting',
            name='app_plugin',
            field=models.ForeignKey(
                help_text='App to which the setting belongs',
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='settings',
                to='djangoplugins.plugin',
            ),
        ),
        migrations.AddField(
            model_name='project',
            name='full_title',
            field=models.CharField(
                help_text='Full project title with parent path (auto-generated)',
                max_length=4096,
                null=True,
            ),
        ),
        migrations.RunPython(
            code=populate_project_full_title,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AlterField(
            model_name='appsetting',
            name='value_json',
            field=models.JSONField(
                default=dict,
                help_text='Optional JSON value for the setting',
                null=True,
            ),
        ),
        migrations.AddField(
            model_name='project',
            name='public_guest_access',
            field=models.BooleanField(
                default=False,
                help_text='Allow public guest access for the project, also including unauthenticated users if allowed on the site',
            ),
        ),
        migrations.AddField(
            model_name='project',
            name='has_public_children',
            field=models.BooleanField(
                default=False,
                help_text='Whether project has children with public access (auto-generated)',
            ),
        ),
        migrations.RunPython(
            code=populate_public_children,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RemoveField(
            model_name='project',
            name='submit_status',
        ),
        migrations.AddField(
            model_name='role',
            name='rank',
            field=models.IntegerField(
                default=0, help_text='Role rank for determining role hierarchy'
            ),
        ),
        migrations.RunPython(
            code=set_role_ranks,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AddField(
            model_name='project',
            name='archive',
            field=models.BooleanField(
                default=False, help_text='Project is archived (read-only)'
            ),
        ),
        migrations.RunPython(
            code=migrate_project_stars,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.DeleteModel(
            name='ProjectUserTag',
        ),
        migrations.RunPython(
            code=delete_category_app_settings,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AddField(
            model_name='role',
            name='project_types',
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=64),
                default=projectroles.models.get_role_project_type_default,
                help_text='Allowed project types for the role',
                size=None,
            ),
        ),
        migrations.AlterField(
            model_name='roleassignment',
            name='project',
            field=models.ForeignKey(
                help_text='Project in which role is assigned',
                on_delete=django.db.models.deletion.CASCADE,
                related_name='local_roles',
                to='projectroles.project',
            ),
        ),
        migrations.RunPython(
            code=populate_finder_role,
            reverse_code=reverse_finder_role,
        ),
        migrations.CreateModel(
            name='SODARUserAdditionalEmail',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'email',
                    models.EmailField(
                        help_text='Email address', max_length=254
                    ),
                ),
                (
                    'verified',
                    models.BooleanField(
                        default=False, help_text='Email verification status'
                    ),
                ),
                (
                    'secret',
                    models.CharField(
                        help_text='Secret token for email verification',
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    'date_created',
                    models.DateTimeField(
                        auto_now_add=True, help_text='DateTime of creation'
                    ),
                ),
                (
                    'date_modified',
                    models.DateTimeField(
                        auto_now=True, help_text='DateTime of last modification'
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='SODARUserAdditionalEmail SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='User for whom the email is assigned',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='additional_emails',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': ['user__username', 'email'],
                'unique_together': {('user', 'email')},
            },
        ),
        migrations.RunPython(
            code=populate_additional_email_model,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AddField(
            model_name='remotesite',
            name='owner_modifiable',
            field=models.BooleanField(
                default=True,
                help_text='Allow owners and delegates to modify project access for this site',
            ),
        ),
        migrations.AlterField(
            model_name='remotesite',
            name='user_display',
            field=models.BooleanField(
                default=True, help_text='Display site to users'
            ),
        ),
        migrations.AlterField(
            model_name='appsetting',
            name='value',
            field=models.CharField(
                blank=True, help_text='Value of the setting', null=True
            ),
        ),
    ]



# Generated by Django 4.2.14 on 2024-07-16 09:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    replaces = [
        ('sodarcache', '0001_initial'),
        ('sodarcache', '0002_make_user_optional'),
        ('sodarcache', '0003_add_index'),
        ('sodarcache', '0004_update_jsonfield'),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('projectroles', '0006_add_remote_projects'),
    ]

    operations = [
        migrations.CreateModel(
            name='JSONCacheItem',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'app_name',
                    models.CharField(help_text='App name', max_length=255),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='Name or title of the item given by the data '
                        'setting app',
                        max_length=255,
                    ),
                ),
                (
                    'date_modified',
                    models.DateTimeField(
                        auto_now_add=True, help_text='DateTime of the update'
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='Item SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'data',
                    models.JSONField(
                        default=dict, help_text='Cached data as JSON'
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        blank=True,
                        help_text='Project in which the item belongs '
                        '(optional)',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='cached_items',
                        to='projectroles.project',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        blank=True,
                        help_text='User who updated the item (optional)',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'abstract': False,
                'unique_together': {('project', 'app_name', 'name')},
                'indexes': [
                    models.Index(
                        fields=['project', 'app_name', 'name'],
                        name='sodarcache__project_e54b84_idx',
                    )
                ],
            },
        ),
    ]



"""
Models and model helper code provided by the bgjobs app.

The central class is BackgroundJob that stores the core information about a
background job.  You can "subclass" this class by creating your own models in
your apps and having a bg_job field referencing BackgroundJob
through a OneToOneField.

Further, the BackgroundJobLogEntry model allows to manage background log
entries for your background jobs.  Use the JobModelMessageMixin for adding
helper functions for applying state changes and adding log messages.
"""

import contextlib
import uuid as uuid_object

from django.conf import settings
from django.db import models

# Projectroles dependency
from projectroles.models import Project


# Access Django user model
AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')

# Levels to use in BackgroundJobLogEntry
LOG_LEVEL_DEBUG = 'debug'
LOG_LEVEL_INFO = 'info'
LOG_LEVEL_WARNING = 'warning'
LOG_LEVEL_ERROR = 'error'

# Choices to use in CharField for log level
LOG_LEVEL_CHOICES = (
    (LOG_LEVEL_DEBUG, 'debug'),
    (LOG_LEVEL_INFO, 'info'),
    (LOG_LEVEL_WARNING, 'warning'),
    (LOG_LEVEL_ERROR, 'error'),
)

# The possible states of BackgroundJob objects and their labels
JOB_STATE_INITIAL = 'initial'
JOB_STATE_RUNNING = 'running'
JOB_STATE_DONE = 'done'
JOB_STATE_FAILED = 'failed'

#: Choices to use in the CharField
JOB_STATE_CHOICES = (
    (JOB_STATE_INITIAL, 'initial'),
    ('running', 'running'),
    ('done', 'done'),
    ('failed', 'failed'),
)


class BackgroundJob(models.Model):
    """Common background job information."""

    #: DateTime of creation
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of creation'
    )
    #: DateTime of last modification
    date_modified = models.DateTimeField(
        auto_now=True, help_text='DateTime of last modification'
    )
    #: The UUID for this job
    sodar_uuid = models.UUIDField(
        default=uuid_object.uuid4, unique=True, help_text='BG Job SODAR UUID'
    )
    #: The project this job belongs to. Set to None for site-wide jobs
    project = models.ForeignKey(
        Project,
        null=True,
        help_text='Project in which this objects belongs',
        on_delete=models.CASCADE,
    )
    #: The user initiating the job
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        null=False,
        related_name='background_jobs',
        on_delete=models.CASCADE,
    )
    #: Specializing string of the job
    job_type = models.CharField(
        max_length=512, null=False, help_text='Type of the job'
    )
    #: A human-readable name for this job
    name = models.CharField(max_length=512)
    #: An optional, extend description for this job
    description = models.TextField()
    #: The job status
    status = models.CharField(
        max_length=50, choices=JOB_STATE_CHOICES, default=JOB_STATE_INITIAL
    )

    class Meta:
        ordering = ["-date_created"]

    def get_human_readable_type(self) -> str:
        """
        Also implement in your sub classes to show human-readable type in the
        views.
        """
        return '(generic job)'

    def add_log_entry(
        self, message: str, level: str = LOG_LEVEL_INFO
    ) -> 'BackgroundJobLogEntry':
        """Add and return a new BackgroundJobLogEntry."""
        return self.log_entries.create(level=level, message=message)

    def __str__(self):
        return self.name


class BackgroundJobLogEntry(models.Model):
    """Log entry for background job"""

    #: Creation time of log entry
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of creation'
    )

    #: The BackgroundJob that the log entry is for
    job = models.ForeignKey(
        BackgroundJob,
        related_name='log_entries',
        help_text='Owning background job',
        on_delete=models.CASCADE,
    )

    #: The entry's log level
    level = models.CharField(
        max_length=50, choices=LOG_LEVEL_CHOICES, help_text='Level of log entry'
    )
    #: The message contained by the log entry
    message = models.TextField(help_text="Log level's message")

    class Meta:
        ordering = ['date_created']


class JobModelMessageMixin:
    """
    Mixin with shortcuts for marking job state and adding log entry.

    Use this in your BackgroundJob "subclasses" (sub classing meaning
    OneToOneField specializations).
    """

    task_desc = None

    @contextlib.contextmanager
    def marks(self):
        """
        Return a context manager that allows to run tasks between start and
        success/error marks.
        """
        self.mark_start()
        try:
            yield
        except Exception as e:
            self.mark_error(f'Error: {e}')
            raise
        else:
            self.mark_success()

    def mark_start(self):
        """Mark the export job as started."""
        self.bg_job.status = JOB_STATE_RUNNING
        self.bg_job.add_log_entry(f'{self.task_desc} started')
        self.bg_job.save()

    def mark_error(self, msg: str):
        """Mark the export job as complete successfully."""
        self.bg_job.status = JOB_STATE_FAILED
        self.bg_job.add_log_entry(f'{self.task_desc} file failed: {msg}')
        self.bg_job.save()

    def mark_success(self):
        """Mark the export job as complete successfully."""
        self.bg_job.status = JOB_STATE_DONE
        self.bg_job.add_log_entry(f'{self.task_desc} succeeded')
        self.bg_job.save()

    def add_log_entry(self, *args, **kwargs) -> BackgroundJobLogEntry:
        """Add a log entry through the related BackgroundJob."""
        return self.bg_job.add_log_entry(*args, **kwargs)



"""UI tests for the userprofile app"""

from django.test import override_settings
from django.urls import reverse

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.forms import APP_SETTING_DISABLE_LABEL
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_models import SODARUserAdditionalEmailMixin
from projectroles.tests.test_ui import UITestBase


app_settings = AppSettingAPI()


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']

# Local constants
APP_NAME_PR = 'projectroles'


@override_settings(AUTH_LDAP_USERNAME_DOMAIN='EXAMPLE')
class TestUserDetailView(SODARUserAdditionalEmailMixin, UITestBase):
    """Tests for UserDetailView"""

    def setUp(self):
        super().setUp()
        # Create users
        self.local_user = self.make_user('local_user', False)
        self.ldap_user = self.make_user('user@EXAMPLE', False)
        self.url = reverse('userprofile:detail')

    def test_update_button(self):
        """Test existence of user update button"""
        expected = [(self.local_user, 1), (self.ldap_user, 0)]
        self.assert_element_count(expected, self.url, 'sodar-user-btn-update')

    def test_update_button_read_only(self):
        """Test existence of user update button with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        expected = [(self.local_user, 0), (self.ldap_user, 0)]
        self.assert_element_count(expected, self.url, 'sodar-user-btn-update')

    def test_settings_button(self):
        """Test existence of settings update button"""
        expected = [
            (self.superuser, 1),
            (self.local_user, 1),
            (self.ldap_user, 1),
        ]
        self.assert_element_count(expected, self.url, 'sodar-user-btn-settings')

    def test_settings_button_read_only(self):
        """Test existence of settings update button with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        expected = [
            (self.superuser, 1),
            (self.local_user, 0),
            (self.ldap_user, 0),
        ]
        self.assert_element_count(expected, self.url, 'sodar-user-btn-settings')

    def test_add_email_button(self):
        """Test existence of add email button"""
        expected = [
            (self.superuser, 1),
            (self.local_user, 1),
            (self.ldap_user, 1),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-user-btn-email-add'
        )

    def test_add_email_button_read_only(self):
        """Test existence of add email button with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        expected = [
            (self.superuser, 1),
            (self.local_user, 0),
            (self.ldap_user, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-user-btn-email-add'
        )

    def test_additional_email_unset(self):
        """Test existence of additional email elements without email"""
        self.assert_element_count(
            [(self.local_user, 0)],
            self.url,
            'sodar-user-email-table-row',
            'class',
        )
        self.assert_element_count(
            [(self.local_user, 0)],
            self.url,
            'sodar-user-email-dropdown',
            'class',
        )
        self.assert_element_exists(
            [self.local_user],
            self.url,
            'sodar-user-email-table-not-found',
            True,
        )

    def test_additional_email_set(self):
        """Test existence of additional email elements with email"""
        self.make_email(self.local_user, 'add1@example.com')
        self.make_email(self.local_user, 'add2@example.com', verified=False)
        # Another user, should not be visible
        self.make_email(self.ldap_user, 'add3@example.com')
        self.assert_element_count(
            [(self.local_user, 2)],
            self.url,
            'sodar-user-email-table-row',
            'class',
        )
        self.assert_element_count(
            [(self.local_user, 2)],
            self.url,
            'sodar-user-email-dropdown',
            'class',
        )
        self.assert_element_exists(
            [self.local_user],
            self.url,
            'sodar-user-email-table-not-found',
            False,
        )

    def test_additional_email_set_read_only(self):
        """Test additional email with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.make_email(self.local_user, 'add1@example.com')
        self.make_email(self.local_user, 'add2@example.com', verified=False)
        self.assert_element_count(
            [(self.local_user, 2)],
            self.url,
            'sodar-user-email-table-row',
            'class',
        )
        self.assert_element_count(
            [(self.local_user, 0)],
            self.url,
            'sodar-user-email-dropdown',
            'class',
        )

    def test_additional_email_set_read_only_superuser(self):
        """Test additional email with site read-only mode as superuser"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.make_email(self.superuser, 'add1@example.com')
        self.make_email(self.superuser, 'add2@example.com', verified=False)
        self.assert_element_count(
            [(self.superuser, 2)],
            self.url,
            'sodar-user-email-table-row',
            'class',
        )
        self.assert_element_count(
            [(self.superuser, 2)],
            self.url,
            'sodar-user-email-dropdown',
            'class',
        )

    @override_settings(PROJECTROLES_SEND_EMAIL=False)
    def test_additional_email_disabled(self):
        """Test existence of email card with PROJECTROLES_SEND_EMAIL=False"""
        self.assert_element_exists(
            [self.local_user],
            self.url,
            'sodar-user-email-card',
            False,
        )

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_additional_email_target(self):
        """Test existence of additional email elements as target site"""
        self.make_email(self.local_user, 'add1@example.com')
        self.make_email(self.local_user, 'add2@example.com', verified=False)
        self.assert_element_exists(
            [self.local_user],
            self.url,
            'sodar-user-btn-email-add',
            False,
        )
        self.assert_element_count(
            [(self.local_user, 2)],
            self.url,
            'sodar-user-email-table-row',
            'class',
        )
        self.assert_element_count(
            [(self.local_user, 0)],
            self.url,
            'sodar-user-email-dropdown',
            'class',
        )
        self.assert_element_exists(
            [self.local_user],
            self.url,
            'sodar-user-email-table-not-found',
            False,
        )


class TestUserAppSettingsView(UITestBase):
    """Tests for UserAppSettingsView"""

    def setUp(self):
        super().setUp()
        # Create users
        self.local_user = self.make_user('local_user', False)
        self.ldap_user = self.make_user('user@EXAMPLE', False)
        self.url = reverse('userprofile:settings_update')

    def test_settings_label_icon(self):
        """Test existence of settings label icon"""
        self.login_and_redirect(self.superuser, self.url)
        WebDriverWait(self.selenium, 15).until(
            lambda x: x.find_element(
                By.CSS_SELECTOR,
                'div[id="div_id_settings.example_project_app.'
                'user_int_setting"] label',
            )
        )
        label = self.selenium.find_element(
            By.CSS_SELECTOR,
            'div[id="div_id_settings.example_project_app.'
            'user_int_setting"] label',
        )
        # Wait before icon is rendered
        WebDriverWait(label, 15).until(
            lambda x: x.find_element(By.TAG_NAME, 'svg')
        )
        icon = label.find_element(By.TAG_NAME, 'svg')
        self.assertTrue(icon.is_displayed())

    def test_global_setting_source(self):
        """Test global user setting on source site"""
        self.login_and_redirect(self.superuser, self.url)
        WebDriverWait(self.selenium, 15).until(
            lambda x: x.find_element(
                By.CSS_SELECTOR,
                'div[id="div_id_settings.projectroles.notify_email_project"]',
            )
        )
        input_elem = self.selenium.find_element(
            By.CSS_SELECTOR,
            'input[id="id_settings.projectroles.notify_email_project"]',
        )
        self.assertIsNone(input_elem.get_attribute('disabled'))
        label = self.selenium.find_element(
            By.CSS_SELECTOR,
            'div[id="div_id_settings.projectroles.notify_email_project"] label',
        )
        self.assertNotIn(APP_SETTING_DISABLE_LABEL, label.text)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_global_setting_target(self):
        """Test global user setting on target site"""
        self.login_and_redirect(self.superuser, self.url)
        WebDriverWait(self.selenium, 15).until(
            lambda x: x.find_element(
                By.CSS_SELECTOR,
                'div[id="div_id_settings.projectroles.notify_email_project"]',
            )
        )
        input_elem = self.selenium.find_element(
            By.CSS_SELECTOR,
            'input[id="id_settings.projectroles.notify_email_project"]',
        )
        self.assertIsNotNone(input_elem.get_attribute('disabled'))
        label = self.selenium.find_element(
            By.CSS_SELECTOR,
            'div[id="div_id_settings.projectroles.notify_email_project"] label',
        )
        self.assertIn(APP_SETTING_DISABLE_LABEL, label.text)



"""Tests for UI view permissions in the timeline app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions import ProjectPermissionTestBase


class TestProjectTimelineView(ProjectPermissionTestBase):
    """Tests for ProjectTimelineView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test ProjectTimelineView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonynomus access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)


class TestSiteTimelineView(ProjectPermissionTestBase):
    """Tests for SiteTimelineView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse('timeline:list_site')

    def test_get(self):
        """Test SiteTimelineView GET"""
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonynomus access"""
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 302)


class TestAdminTimelineView(ProjectPermissionTestBase):
    """Tests for AdminTimelineView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse('timeline:list_admin')

    def test_get(self):
        """Test AdminTimelineView GET"""
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.no_role_users, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)


class TestProjectObjectTimelineView(ProjectPermissionTestBase):
    """Tests for ProjectObjectTimelineView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': 'User',
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test ProjectObjectTimelineView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_event_object_list_anon(self):
        """Test GET with anonynomus access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.non_superusers, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 302)


class TestSiteObjectTimelineView(ProjectPermissionTestBase):
    """Tests for SiteObjectTimelineView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': 'User',
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )

    def test_get(self):
        """Test SiteObjectTimelineView GET"""
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 302)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 302)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 302)



"""SODAR constants definition and helper functions"""

# Global SODAR constants
SODAR_CONSTANTS = {
    # Project roles
    'PROJECT_ROLE_OWNER': 'project owner',
    'PROJECT_ROLE_DELEGATE': 'project delegate',
    'PROJECT_ROLE_CONTRIBUTOR': 'project contributor',
    'PROJECT_ROLE_GUEST': 'project guest',
    'PROJECT_ROLE_VIEWER': 'project viewer',
    'PROJECT_ROLE_FINDER': 'project finder',
    # Project types
    'PROJECT_TYPE_CATEGORY': 'CATEGORY',
    'PROJECT_TYPE_PROJECT': 'PROJECT',
    # App Settings
    'APP_SETTING_SCOPE_PROJECT': 'PROJECT',
    'APP_SETTING_SCOPE_USER': 'USER',
    'APP_SETTING_SCOPE_PROJECT_USER': 'PROJECT_USER',
    'APP_SETTING_SCOPE_SITE': 'SITE',
    'APP_SETTING_TYPE_BOOLEAN': 'BOOLEAN',
    'APP_SETTING_TYPE_INTEGER': 'INTEGER',
    'APP_SETTING_TYPE_JSON': 'JSON',
    'APP_SETTING_TYPE_STRING': 'STRING',
    # RemoteSite mode
    'SITE_MODE_SOURCE': 'SOURCE',
    'SITE_MODE_TARGET': 'TARGET',
    'SITE_MODE_PEER': 'PEER',
    # RemoteProject access types
    'REMOTE_LEVEL_NONE': 'NONE',
    'REMOTE_LEVEL_REVOKED': 'REVOKED',
    'REMOTE_LEVEL_VIEW_AVAIL': 'VIEW_AVAIL',
    'REMOTE_LEVEL_READ_INFO': 'READ_INFO',
    'REMOTE_LEVEL_READ_ROLES': 'READ_ROLES',
    # RemoteSite modes
    'SITE_MODES': ['SOURCE', 'TARGET', 'PEER'],
    # RemoteProject access type legend
    'REMOTE_ACCESS_LEVELS': {
        'NONE': 'No access',
        'REVOKED': 'Revoked access',
        'VIEW_AVAIL': 'View availability',
        'READ_INFO': 'Read information',
        'READ_ROLES': 'Read members',
    },
    # Display names
    'DISPLAY_NAMES': {
        'CATEGORY': {'default': 'category', 'plural': 'categories'},
        'PROJECT': {'default': 'project', 'plural': 'projects'},
    },
    # User types
    'AUTH_TYPE_LOCAL': 'LOCAL',
    'AUTH_TYPE_LDAP': 'LDAP',
    'AUTH_TYPE_OIDC': 'OIDC',
    # System user group
    'SYSTEM_USER_GROUP': 'system',
    'OIDC_USER_GROUP': 'oidc',
    # Project modification
    'PROJECT_ACTION_CREATE': 'CREATE',
    'PROJECT_ACTION_UPDATE': 'UPDATE',
}


def get_sodar_constants(default=False):
    """
    Return SODAR_CONSTANTS from settings if present, else from default
    definition.
    """
    if not default:
        from django.conf import settings

        if hasattr(settings, 'SODAR_CONSTANTS'):
            return settings.SODAR_CONSTANTS
    return SODAR_CONSTANTS



# Generated by Django 4.2.14 on 2024-07-16 09:17

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import markupfield.fields
import uuid


class Migration(migrations.Migration):

    replaces = [
        ('adminalerts', '0001_initial'),
        ('adminalerts', '0002_adminalert_user'),
        ('adminalerts', '0003_auto_20180802_1558'),
        ('adminalerts', '0004_rename_uuid'),
        ('adminalerts', '0005_update_uuid'),
        ('adminalerts', '0006_adminalert_require_auth'),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AdminAlert',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'message',
                    models.CharField(
                        help_text='Alert message to be shown for users',
                        max_length=255,
                    ),
                ),
                (
                    'description',
                    markupfield.fields.MarkupField(
                        blank=True,
                        help_text='Full description of alert (optional, will '
                        'be shown on a separate page)',
                        null=True,
                        rendered_field=True,
                    ),
                ),
                (
                    'date_created',
                    models.DateTimeField(
                        auto_now_add=True, help_text='Alert creation timestamp'
                    ),
                ),
                (
                    'date_expire',
                    models.DateTimeField(
                        help_text='Alert expiration timestamp'
                    ),
                ),
                (
                    'active',
                    models.BooleanField(
                        default=True,
                        help_text='Alert status (for disabling the alert '
                        'before expiration)',
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='Adminalerts SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='Superuser who has set the alert',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='alerts',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    '_description_rendered',
                    models.TextField(editable=False, null=True),
                ),
                (
                    'description_markup_type',
                    models.CharField(
                        choices=[
                            ('', '--'),
                            ('html', 'HTML'),
                            ('plain', 'Plain'),
                            ('markdown', 'Markdown'),
                            ('restructuredtext', 'Restructured Text'),
                        ],
                        default='markdown',
                        editable=False,
                        max_length=30,
                    ),
                ),
                (
                    'require_auth',
                    models.BooleanField(
                        default=True,
                        help_text='Require authorization to view alert',
                    ),
                ),
            ],
        ),
    ]



"""Models for the appalerts app"""

import uuid

from django.core.exceptions import ValidationError
from django.conf import settings
from django.db import models
from djangoplugins.models import Plugin

# Projectroles dependency
from projectroles.models import Project


# Access Django user model
AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


# Local constants
ALERT_LEVELS = ['INFO', 'SUCCESS', 'WARNING', 'DANGER']
ALERT_LEVEL_CHOICES = [(a, a.capitalize()) for a in ALERT_LEVELS]


class AppAlert(models.Model):
    """Class representing an alert directed to a specific user"""

    #: App to which the alert belongs
    app_plugin = models.ForeignKey(
        Plugin,
        null=True,  # Null for projectroles
        related_name='app_alerts',
        help_text='App to which the alert belongs',
        on_delete=models.CASCADE,
    )

    #: Internal alert name string
    alert_name = models.CharField(
        max_length=255, help_text='Internal alert name string'
    )

    #: User who receives the alert
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        null=False,
        related_name='app_alerts',
        help_text='User who receives the alert',
        on_delete=models.CASCADE,
    )

    #: Alert message (may contain HTML)
    message = models.TextField(help_text='Alert message (may contain HTML)')

    #: Alert level
    level = models.CharField(
        max_length=64,
        blank=False,
        null=False,
        default='INFO',
        choices=ALERT_LEVEL_CHOICES,
        help_text='Alert level',
    )

    #: Active status of the alert
    active = models.BooleanField(
        default=True, help_text='Active status of the alert'
    )

    #: URL for the source of the alert (optional)
    url = models.URLField(
        max_length=2000,
        blank=True,
        null=True,
        help_text='URL for the source of the alert (optional)',
    )

    #: Project to which the alert belongs (optional)
    project = models.ForeignKey(
        Project,
        related_name='app_alerts',
        help_text='Project to which the alert belongs (optional)',
        on_delete=models.CASCADE,
        null=True,
    )

    #: DateTime of the alert creation
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of the alert creation'
    )

    #: UUID for the alert
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Alert SODAR UUID'
    )

    def __str__(self):
        return '{} / {} / {}'.format(
            self.app_plugin.name if self.app_plugin else 'projectroles',
            self.alert_name,
            self.user.username,
        )

    def __repr__(self):
        values = (
            self.app_plugin.name if self.app_plugin else 'projectroles',
            self.alert_name,
            self.user.username,
            self.project.title if self.project else 'N/A',
        )
        return 'AppAlert({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Custom validation for AppAlert"""
        self._validate_level()
        super().save(*args, **kwargs)

    def _validate_level(self):
        """Validate level"""
        if self.level not in ALERT_LEVELS:
            raise ValidationError(
                f'Invalid level "{self.level}", valid levels: '
                f'{", ".join(ALERT_LEVELS)}'
            )



"""REST API view permission tests for the timeline app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions_api import ProjectAPIPermissionTestBase

from timeline.tests.test_models import TimelineEventMixin
from timeline.tests.test_views_api import (
    EVENT_NAME,
    EVENT_DESC,
    EXTRA_DATA,
)
from timeline.views_api import (
    TIMELINE_API_MEDIA_TYPE,
    TIMELINE_API_DEFAULT_VERSION,
)


class TimelineAPIPermissionTestBase(ProjectAPIPermissionTestBase):
    """Base class for timeline REST API view permission tests"""

    media_type = TIMELINE_API_MEDIA_TYPE
    api_version = TIMELINE_API_DEFAULT_VERSION


class TestProjectTimelineEventListAPIView(TimelineAPIPermissionTestBase):
    """Tests for ProjectTimelineEventListAPIView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'timeline:api_list_project',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test ProjectTimelineEventListAPIView GET"""

        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)


class TestSiteTimelineEventListAPIView(TimelineAPIPermissionTestBase):
    """Tests for SiteTimelineEventListAPIView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('timeline:api_list_site')

    def test_get(self):
        """Test SiteTimelineEventListAPIView GET"""
        self.assert_response_api(self.url, self.auth_users, 200)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.auth_users, 200, knox=True)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.auth_users, 200)
        self.assert_response_api(self.url, self.anonymous, 401)


class TestTimelineEventRetrieveAPIView(
    TimelineEventMixin, TimelineAPIPermissionTestBase
):
    """Tests for TimelineEventRetrieveAPIView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            project=self.project,
            app='projectroles',
            user=self.user_owner,
            event_name=EVENT_NAME,
            description=EVENT_DESC,
            classified=False,
            extra_data=EXTRA_DATA,
        )
        self.url = reverse(
            'timeline:api_retrieve',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )
        self.good_users_project = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users_project = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get_project_event(self):
        """Test TimelineEventRetrieveAPIView GET with project event"""
        self.assert_response_api(self.url, self.good_users_project, 200)
        self.assert_response_api(self.url, self.bad_users_project, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(
            self.url, self.good_users_project, 200, knox=True
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_project_event_anon(self):
        """Test GET with project event and anonymous access"""
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_project_event_read_only(self):
        """Test GET with project event and site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users_project, 200)
        self.assert_response_api(self.url, self.bad_users_project, 403)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_project_event_classified(self):
        """Test GET with classified project event"""
        self.event.classified = True
        self.event.save()
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]
        self.assert_response_api(self.url, good_users, 200)
        self.assert_response_api(self.url, bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, good_users, 200, knox=True)

    def test_get_site_event(self):
        """Test GET with site-wide event"""
        self.event.project = None
        self.event.save()
        self.assert_response_api(self.url, self.auth_users, 200)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.auth_users, 200, knox=True)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_site_event_anon(self):
        """Test GET with site event and anonymous access"""
        self.event.project = None
        self.event.save()
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_site_event_read_only(self):
        """Test GET with site-wide event and site read-only mode"""
        self.set_site_read_only()
        self.event.project = None
        self.event.save()
        self.assert_response_api(self.url, self.auth_users, 200)
        self.assert_response_api(self.url, self.anonymous, 401)

    def test_get_site_event_classified(self):
        """Test GET with classified site event"""
        self.event.project = None
        self.event.classified = True
        self.event.save()
        self.assert_response_api(self.url, self.superuser, 200)
        self.assert_response_api(self.url, self.auth_non_superusers, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.superuser, 200, knox=True)



"""Plugins for the Timeline app"""

from typing import Optional

# Projectroles dependency
from projectroles.models import SODARUser, SODAR_CONSTANTS
from projectroles.plugins import (
    ProjectAppPluginPoint,
    BackendPluginPoint,
    SiteAppPluginPoint,
    PluginSearchResult,
)
from projectroles.utils import get_display_name

from timeline.api import TimelineAPI
from timeline.models import TimelineEvent
from timeline.urls import urls_ui_project, urls_ui_site, urls_ui_admin


# Local constants
STATS_DESC_USER_COUNT = 'Amount of users who have initiated events'


class ProjectAppPlugin(ProjectAppPluginPoint):
    """Plugin for registering app with Projectroles"""

    # Properties required by django-plugins ------------------------------

    #: Name (used as plugin ID)
    name = 'timeline'

    #: Title (used in templates)
    title = 'Timeline'

    #: UI URLs
    urls = urls_ui_project

    # Properties defined in ProjectAppPluginPoint -----------------------

    #: Iconify icon
    icon = 'mdi:clock-time-eight'

    #: Entry point URL ID (must take project sodar_uuid as "project" argument)
    entry_point_url_id = 'timeline:list_project'

    #: Description string
    description = 'Timeline of {} events'.format(
        get_display_name(SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'])
    )

    #: Required permission for accessing the app
    app_permission = 'timeline.view_timeline'

    #: Enable or disable general search from project title bar
    search_enable = True

    #: List of search object types for the app
    search_types = ['timeline']

    #: Search results template
    search_template = 'timeline/_search_results.html'

    #: App card template for the project details page
    details_template = 'timeline/_details_card.html'

    #: App card title for the project details page
    details_title = 'Timeline Overview'

    #: Position in plugin ordering
    plugin_ordering = 40

    #: Display application for categories in addition to projects
    category_enable = True

    #: Names of plugin specific Django settings to display in siteinfo
    info_settings = ['TIMELINE_PAGINATION', 'TIMELINE_SEARCH_LIMIT']

    @classmethod
    def _check_permission(cls, user: SODARUser, event: TimelineEvent) -> bool:
        """Check if user has permission to view event"""
        if event.project and event.classified:
            return user.has_perm(
                'timeline.view_classified_event', event.project
            )
        elif event.project:
            return user.has_perm('timeline.view_timeline', event.project)
        elif event.classified:
            return user.has_perm('timeline.view_classified_site_event')
        return user.has_perm('timeline.view_site_timeline')

    def get_statistics(self) -> dict:
        return {
            'event_count': {
                'label': 'Events',
                'value': TimelineEvent.objects.all().count(),
            },
            'user_count': {
                'label': 'Users',
                'description': STATS_DESC_USER_COUNT,
                'value': TimelineEvent.objects.exclude(user__isnull=True)
                .values('user')
                .distinct()
                .count(),
            },
        }

    def search(
        self,
        search_terms: list[str],
        user: SODARUser,
        search_type: Optional[str] = None,
        keywords: Optional[list[str]] = None,
    ) -> list[PluginSearchResult]:
        """
        Return app items based on one or more search terms, user, optional type
        and optional keywords.

        :param search_terms: Search terms to be joined with the OR operator
                             (list of strings)
        :param user: User object for user initiating the search
        :param search_type: String
        :param keywords: List (optional)
        :return: List of PluginSearchResult objects
        """
        items = []
        if not search_type or search_type == 'timeline':
            events = list(TimelineEvent.objects.find(search_terms, keywords))
            items = [e for e in events if self._check_permission(user, e)]
        ret = PluginSearchResult(
            category='all',
            title='Timeline Events',
            search_types=['timeline'],
            items=items,
        )
        return [ret]


class BackendPlugin(BackendPluginPoint):
    """Plugin for registering backend app with Projectroles"""

    #: Name (used as plugin ID)
    name = 'timeline_backend'

    #: Title (used in templates)
    title = 'Timeline Backend'

    #: Iconify icon
    icon = 'mdi:clock-time-eight-outline'

    #: Description string
    description = 'Timeline backend for modifying events'

    def get_api(self, **kwargs) -> TimelineAPI:
        """Return API entry point object."""
        return TimelineAPI()


class SiteAppPlugin(SiteAppPluginPoint):
    """Projectroles plugin for registering the app"""

    #: Name (used as plugin ID)
    name = 'timeline_site'

    #: Title (used in templates)
    title = 'Site-Wide Events'

    #: UI URLs
    urls = urls_ui_site

    #: Iconify icon
    icon = 'mdi:clock-time-eight-outline'

    #: Description string
    description = 'Timeline of Site-Wide Events'

    #: Entry point URL ID
    entry_point_url_id = 'timeline:list_site'

    #: Required permission for displaying the app
    app_permission = 'timeline.view_site_timeline'


class AdminSiteAppPlugin(SiteAppPluginPoint):
    """Projectroles plugin for registering the app"""

    #: Name (used as plugin ID)
    name = 'timeline_site_admin'

    #: Title (used in templates)
    title = 'All Timeline Events'

    #: UI URLs
    urls = urls_ui_admin

    #: Iconify icon
    # icon = 'mdi:clock-star-four-points-outline'
    icon = 'mdi:web-clock'

    #: Description string
    description = 'Admin view for all timeline events on the site'

    #: Entry point URL ID
    entry_point_url_id = 'timeline:list_admin'

    #: Required permission for displaying the app
    app_permission = 'timeline.view_site_admin'



"""Tests for management commands in the projectroles app"""

import os

from tempfile import NamedTemporaryFile
from typing import Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import mail
from django.core.management import call_command
from django.forms.models import model_to_dict
from django.test import override_settings
from djangoplugins.models import Plugin

from test_plus.test import TestCase

# Appalerts dependency
from appalerts.models import AppAlert

# Timeline dependency
from timeline.models import TimelineEvent

from projectroles.app_settings import AppSettingAPI
from projectroles.management.commands.addremotesite import (
    Command as AddRemoteSiteCommand,
)
from projectroles.management.commands.batchupdateroles import (
    Command as BatchUpdateRolesCommand,
)
from projectroles.management.commands.blockprojectaccess import (
    Command as BlockProjectAccessCommand,
    INVALID_PROJECT_TYPE_MSG,
    PROJECT_NOT_FOUND_MSG,
    INVALID_MODE_MSG,
    MODE_CONVERT_ERR_MSG,
)
from projectroles.management.commands.cleanappsettings import (
    LOG_NONE_LABEL,
    START_MSG,
    END_MSG,
    DEF_NOT_FOUND_MSG,
    ALLOWED_TYPES_MSG,
    DELETE_PREFIX_MSG,
    DELETE_PROJECT_TYPE_MSG,
    DELETE_SCOPE_MSG,
)
from projectroles.management.commands.createdevusers import (
    DEV_USER_NAMES,
    DEFAULT_PASSWORD,
)
from projectroles.management.commands.syncgroups import (
    Command as SyncGroupsCommand,
)

from projectroles.models import (
    RoleAssignment,
    ProjectInvite,
    RemoteSite,
    AppSetting,
    SODAR_CONSTANTS,
)
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    ProjectInviteMixin,
    AppSettingMixin,
    RemoteSiteMixin,
    RemoteProjectMixin,
)
from projectroles.utils import build_secret


app_settings = AppSettingAPI()
User = get_user_model()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_PEER = SODAR_CONSTANTS['SITE_MODE_PEER']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
REMOTE_LEVEL_READ_ROLES = SODAR_CONSTANTS['REMOTE_LEVEL_READ_ROLES']
SYSTEM_USER_GROUP = SODAR_CONSTANTS['SYSTEM_USER_GROUP']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
APP_NAME = 'projectroles'
EXAMPLE_APP_NAME = 'example_project_app'
LOGGER_PREFIX = 'projectroles.management.commands.'
CLEAN_LOG_PREFIX = 'INFO:projectroles.management.commands.cleanappsettings:'
CUSTOM_PASSWORD = 'custompass'
REMOTE_SITE_NAME = 'Test Site'
REMOTE_SITE_URL = 'https://example.com'
REMOTE_SITE_URL2 = 'https://another.site.org'
REMOTE_SITE_SECRET = build_secret(32)
LDAP_DOMAIN = 'EXAMPLE'
INVALID_UUID = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'


# Mixins -----------------------------------------------------------------------


class BatchUpdateRolesMixin:
    """
    Helpers for batchupdateroles testing.

    NOTE: May be needed by external repos for testing, do not remove!
    """

    file = None

    def write_file(self, data: Union[str, list]):
        """Write data to temporary CSV file"""
        if not isinstance(data[0], list):
            data = [data]
        self.file.write(
            bytes('\n'.join([';'.join(r) for r in data]), encoding='utf-8')
        )
        self.file.close()


# Tests ------------------------------------------------------------------------


class TestAddRemoteSite(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    RemoteSiteMixin,
    TestCase,
):
    """Tests for addremotesite command"""

    def setUp(self):
        self.user = self.make_user('superuser')
        self.user.is_superuser = True
        self.user.save()
        self.command = AddRemoteSiteCommand()
        self.options = {
            'name': REMOTE_SITE_NAME,
            'url': REMOTE_SITE_URL,
            'mode': SITE_MODE_TARGET,
            'description': '',
            'secret': REMOTE_SITE_SECRET,
            'user_display': True,
            'owner_modifiable': False,
            'suppress_error': False,
        }

    def test_add(self):
        """Test adding new remote site"""
        tl_kwargs = {'event_name': 'target_site_create'}
        self.assertEqual(RemoteSite.objects.count(), 0)
        self.assertEqual(TimelineEvent.objects.filter(**tl_kwargs).count(), 0)
        self.command.handle(**self.options)

        self.assertEqual(RemoteSite.objects.count(), 1)
        site = RemoteSite.objects.first()
        expected = {
            'id': site.pk,
            'name': REMOTE_SITE_NAME,
            'url': REMOTE_SITE_URL,
            'mode': SITE_MODE_TARGET,
            'description': '',
            'secret': REMOTE_SITE_SECRET,
            'sodar_uuid': site.sodar_uuid,
            'user_display': True,
            'owner_modifiable': False,
        }
        self.assertEqual(model_to_dict(site), expected)
        self.assertEqual(TimelineEvent.objects.filter(**tl_kwargs).count(), 1)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_add_source_on_target(self):
        """Test adding source site on target site"""
        tl_kwargs = {'event_name': 'source_site_create'}
        self.options['mode'] = SITE_MODE_SOURCE
        self.assertEqual(RemoteSite.objects.count(), 0)
        self.assertEqual(TimelineEvent.objects.filter(**tl_kwargs).count(), 0)
        self.command.handle(**self.options)
        self.assertEqual(RemoteSite.objects.count(), 1)
        self.assertEqual(TimelineEvent.objects.filter(**tl_kwargs).count(), 1)

    def test_add_source_on_source(self):
        """Test adding source site on source site (should fail)"""
        self.options['mode'] = SITE_MODE_SOURCE
        self.assertEqual(RemoteSite.objects.count(), 0)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 0)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_add_target_on_target(self):
        """Test adding target site on target site (should fail)"""
        self.assertEqual(RemoteSite.objects.count(), 0)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 0)

    def test_add_peer_on_source(self):
        """Test adding peer site on source site (should fail)"""
        self.options['mode'] = SITE_MODE_PEER
        self.assertEqual(RemoteSite.objects.count(), 0)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 0)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_add_peer_on_target(self):
        """Test adding peer site on target site (should fail)"""
        self.options['mode'] = SITE_MODE_PEER
        self.assertEqual(RemoteSite.objects.count(), 0)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 0)

    def test_add_second_target(self):
        """Test adding second target site"""
        self.make_site('Existing Site', REMOTE_SITE_URL2)
        self.assertEqual(RemoteSite.objects.count(), 1)
        self.command.handle(**self.options)
        self.assertEqual(RemoteSite.objects.count(), 2)

    def test_add_second_target_existing_name(self):
        """Test adding second target site with existing name (should fail)"""
        self.make_site(REMOTE_SITE_NAME, REMOTE_SITE_URL2)
        self.assertEqual(RemoteSite.objects.count(), 1)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 1)

    def test_add_second_target_existing_url(self):
        """Test adding second target site with existing URL (should fail)"""
        self.make_site('Existing Site', REMOTE_SITE_URL)
        self.assertEqual(RemoteSite.objects.count(), 1)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 1)

    def test_add_second_target_existing_url_suppress(self):
        """Test adding second target site with suppressed error"""
        self.options['suppress_error'] = True
        self.make_site('Existing Site', REMOTE_SITE_URL)
        self.assertEqual(RemoteSite.objects.count(), 1)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 0)  # No error
        self.assertEqual(RemoteSite.objects.count(), 1)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_add_second_source(self):
        """Test adding second source site (should fail)"""
        self.make_site('Existing Site', REMOTE_SITE_URL2, mode=SITE_MODE_SOURCE)
        self.assertEqual(RemoteSite.objects.count(), 1)
        with self.assertRaises(SystemExit) as cm:
            self.command.handle(**self.options)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(RemoteSite.objects.count(), 1)


class TestBatchUpdateRoles(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    ProjectInviteMixin,
    BatchUpdateRolesMixin,
    TestCase,
):
    """Tests for batchupdateroles command"""

    def setUp(self):
        super().setUp()
        # Init roles
        self.init_roles()
        # Init users
        self.user_owner = self.make_user('owner')
        self.user_owner.email = 'owner_user@example.com'
        self.user_owner.save()
        self.user_owner_cat = self.make_user('owner_cat')
        self.user_owner_cat.email = 'cat_owner_user@example.com'
        self.user_owner_cat.save()

        # Init projects
        self.category = self.make_project(
            'top_category', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self.make_assignment(
            self.category, self.user_owner_cat, self.role_owner
        )
        self.project = self.make_project(
            'sub_project', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        # Save UUIDs for easy access
        self.category_uuid = str(self.category.sodar_uuid)
        self.project_uuid = str(self.project.sodar_uuid)

        # Init command class
        self.command = BatchUpdateRolesCommand()
        # Init file
        self.file = NamedTemporaryFile(delete=False)

    def tearDown(self):
        if self.file:
            os.remove(self.file.name)
        super().tearDown()

    def test_invite(self):
        """Test inviting a single user via email to project"""
        email = 'new@example.com'
        self.assertEqual(ProjectInvite.objects.count(), 0)

        self.write_file([self.project_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 1)
        invite = ProjectInvite.objects.first()
        self.assertEqual(invite.email, email)
        self.assertEqual(invite.project, self.project)
        self.assertEqual(invite.role, self.role_guest)
        self.assertEqual(invite.issuer, self.user_owner)
        self.assertEqual(len(mail.outbox), 1)

    def test_invite_existing(self):
        """Test inviting a user when they already have an active invite"""
        email = 'new@example.com'
        self.make_invite(email, self.project, self.role_guest, self.user_owner)
        self.assertEqual(ProjectInvite.objects.count(), 1)

        self.write_file([self.project_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_multi_user(self):
        """Test inviting multiple users"""
        email = 'new@example.com'
        email2 = 'new2@example.com'
        self.assertEqual(ProjectInvite.objects.count(), 0)

        fd = [
            [self.project_uuid, email, PROJECT_ROLE_GUEST],
            [self.project_uuid, email2, PROJECT_ROLE_GUEST],
        ]
        self.write_file(fd)
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 2)
        self.assertEqual(len(mail.outbox), 2)

    def test_invite_multi_project(self):
        """Test inviting user to multiple projects"""
        email = 'new@example.com'
        self.assertEqual(ProjectInvite.objects.count(), 0)

        fd = [
            [self.category_uuid, email, PROJECT_ROLE_GUEST],
            [self.project_uuid, email, PROJECT_ROLE_GUEST],
        ]
        self.write_file(fd)
        # NOTE: Using user_owner_cat as they have perms for both projects
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner_cat.username}
        )

        self.assertEqual(
            ProjectInvite.objects.filter(project=self.category).count(), 1
        )
        self.assertEqual(
            ProjectInvite.objects.filter(project=self.project).count(), 1
        )
        self.assertEqual(len(mail.outbox), 2)

    def test_invite_owner(self):
        """Test inviting an owner (should fail)"""
        self.write_file(
            [self.project_uuid, 'new@example.com', PROJECT_ROLE_OWNER]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_delegate(self):
        """Test inviting a delegate"""
        self.write_file(
            [self.project_uuid, 'new@example.com', PROJECT_ROLE_DELEGATE]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )
        self.assertEqual(ProjectInvite.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 1)

    @override_settings(PROJECTROLES_DELEGATE_LIMIT=2)
    def test_invite_delegate_no_perms(self):
        """Test inviting a delegate without perms (should fail)"""
        user_delegate = self.make_user('delegate')
        self.make_assignment(self.project, user_delegate, self.role_delegate)
        self.write_file(
            [self.project_uuid, 'new@example.com', PROJECT_ROLE_DELEGATE]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': user_delegate.username}
        )
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_delegate_limit(self):
        """Test inviting a delegate with limit reached (should fail)"""
        user_delegate = self.make_user('delegate')
        self.make_assignment(self.project, user_delegate, self.role_delegate)
        self.write_file(
            [self.project_uuid, 'new@example.com', PROJECT_ROLE_DELEGATE]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add(self):
        """Test adding role to user already in system"""
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        self.assertEqual(ProjectInvite.objects.count(), 0)

        self.write_file([self.project_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as = RoleAssignment.objects.get(
            project=self.project, user=user_new
        )
        self.assertEqual(role_as.role, self.role_guest)
        self.assertEqual(len(mail.outbox), 1)

    def test_role_update(self):
        """Test updating an existing role for user"""
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        role_as = self.make_assignment(self.project, user_new, self.role_guest)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            1,
        )

        self.write_file([self.project_uuid, email, PROJECT_ROLE_CONTRIBUTOR])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as.refresh_from_db()
        self.assertEqual(role_as.role, self.role_contributor)
        self.assertEqual(len(mail.outbox), 1)

    def test_role_update_owner(self):
        """Test updating the role of current owner (should fail)"""
        email = 'owner@example.com'
        self.user_owner.email = email
        self.user_owner.save()

        self.write_file([self.project_uuid, email, PROJECT_ROLE_CONTRIBUTOR])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.owner_as.refresh_from_db()
        self.assertEqual(self.owner_as.role, self.role_owner)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=self.user_owner
            ).count(),
            1,
        )
        self.assertEqual(len(mail.outbox), 0)

    @override_settings(PROJECTROLES_DELEGATE_LIMIT=2)
    def test_role_update_delegate_no_perms(self):
        """Test updating a delegate role without perms (should fail)"""
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        role_as = self.make_assignment(self.project, user_new, self.role_guest)
        user_delegate = self.make_user('delegate')
        self.make_assignment(self.project, user_delegate, self.role_delegate)

        self.write_file([self.project_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': user_delegate.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as.refresh_from_db()
        self.assertEqual(role_as.role, self.role_guest)
        self.assertEqual(len(mail.outbox), 0)

    def test_role_update_delegate_limit(self):
        """Test updating a delegate role with limit reached (should fail)"""
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        role_as = self.make_assignment(self.project, user_new, self.role_guest)
        user_delegate = self.make_user('delegate')
        self.make_assignment(self.project, user_delegate, self.role_delegate)

        self.write_file([self.project_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as.refresh_from_db()
        self.assertEqual(role_as.role, self.role_guest)
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add_inherited_owner(self):
        """Test adding role with inherited owner rank (should fail)"""
        self.write_file(
            [
                self.project_uuid,
                self.user_owner_cat.email,
                PROJECT_ROLE_CONTRIBUTOR,
            ]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=self.user_owner_cat
            ).count(),
            0,
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add_inherited_higher(self):
        """Test adding role with inherited role of higher rank"""
        user_new = self.make_user('user_new')
        user_new.email = 'user_new@example.com'
        self.make_assignment(self.category, user_new, self.role_guest)

        self.write_file(
            [
                self.project_uuid,
                user_new.email,
                PROJECT_ROLE_CONTRIBUTOR,
            ]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            0,
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add_inherited_equal(self):
        """Test adding role with inherited role of equal rank (should fail)"""
        user_new = self.make_user('user_new')
        user_new.email = 'user_new@example.com'
        self.make_assignment(self.category, user_new, self.role_contributor)

        self.write_file(
            [
                self.project_uuid,
                user_new.email,
                PROJECT_ROLE_CONTRIBUTOR,
            ]
        )
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            0,
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add_inherited_lower(self):
        """Test adding role with inherited role of lower rank (should fail)"""
        user_new = self.make_user('user_new')
        user_new.email = 'user_new@example.com'
        self.make_assignment(self.category, user_new, self.role_contributor)

        self.write_file([self.project_uuid, user_new.email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            0,
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_and_update(self):
        """Test inviting and updating in one command"""
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        email2 = 'new2@example.com'
        self.assertEqual(ProjectInvite.objects.count(), 0)

        fd = [
            [self.project_uuid, email, PROJECT_ROLE_GUEST],
            [self.project_uuid, email2, PROJECT_ROLE_GUEST],
        ]
        self.write_file(fd)
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        self.assertEqual(ProjectInvite.objects.count(), 1)
        self.assertIsNotNone(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            1,
        )
        self.assertEqual(len(mail.outbox), 2)

    def test_command_no_issuer(self):
        """Test invite without issuer (should go under admin)"""
        admin = self.make_user(settings.PROJECTROLES_DEFAULT_ADMIN)
        admin.is_superuser = True
        admin.save()
        email = 'new@example.com'
        self.write_file([self.project_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(**{'file': self.file.name, 'issuer': None})
        invite = ProjectInvite.objects.first()
        self.assertEqual(invite.issuer, admin)

    def test_command_no_perms(self):
        """Test invite with issuer who lacks permissions for a project"""
        issuer = self.make_user('issuer')
        email = 'new@example.com'
        self.write_file([self.project_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': issuer.username}
        )
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)


class TestBlockProjectAccess(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    TestCase,
):
    """Tests for blockprojectaccess command"""

    def _assert_setting(self, value: bool = False):
        """
        Assert project_access_block setting value.

        :param value: Expected value (bool)
        """
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'project_access_block', project=self.project
            ),
            value,
        )

    def setUp(self):
        super().setUp()
        # Init roles
        self.init_roles()
        # Init users
        self.user_owner = self.make_user('owner')
        self.user_owner.email = 'owner_user@example.com'
        self.user_owner.save()
        self.user_owner_cat = self.make_user('owner_cat')
        self.user_owner_cat.email = 'cat_owner_user@example.com'
        self.user_owner_cat.save()
        # Init projects
        self.category = self.make_project(
            'top_category', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self.make_assignment(
            self.category, self.user_owner_cat, self.role_owner
        )
        self.project = self.make_project(
            'sub_project', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        # Save UUIDs for easy access
        self.category_uuid = str(self.category.sodar_uuid)
        self.project_uuid = str(self.project.sodar_uuid)
        # Init command class
        self.command = BlockProjectAccessCommand()
        self.options = {'project': str(self.project.sodar_uuid), 'mode': '1'}
        self.logger_name = LOGGER_PREFIX + 'blockprojectaccess'

    def test_command_set(self):
        """Test blockprojectaccess to set value"""
        self._assert_setting(False)
        self.command.handle(**self.options)
        self._assert_setting(True)

    def test_command_unset(self):
        """Test blockprojectaccess to unset value"""
        app_settings.set(
            APP_NAME, 'project_access_block', True, project=self.project
        )
        self._assert_setting(True)
        self.options['mode'] = '0'
        self.command.handle(**self.options)
        self._assert_setting(False)

    def test_command_same_value(self):
        """Test blockprojectaccess with same value as already set"""
        self._assert_setting(False)
        self.options['mode'] = '0'
        self.command.handle(**self.options)
        self._assert_setting(False)

    def test_command_category(self):
        """Test blockprojectaccess with category (should fail)"""
        self.options['project'] = str(self.category.sodar_uuid)
        with self.assertLogs(self.logger_name, 'ERROR') as cm:
            with self.assertRaises(SystemExit):
                self.command.handle(**self.options)
        self.assertIn(INVALID_PROJECT_TYPE_MSG, cm.output[0])

    def test_command_invalid_uuid(self):
        """Test blockprojectaccess with invalid UUID (should fail)"""
        self.options['project'] = INVALID_UUID
        with self.assertLogs(self.logger_name, 'ERROR') as cm:
            with self.assertRaises(SystemExit):
                self.command.handle(**self.options)
        self.assertIn(
            PROJECT_NOT_FOUND_MSG.format(project_uuid=INVALID_UUID),
            cm.output[0],
        )

    def test_command_value_out_of_range(self):
        """Test blockprojectaccess with value out of range (should fail)"""
        self.options['mode'] = '2'
        with self.assertLogs(self.logger_name, 'ERROR') as cm:
            with self.assertRaises(SystemExit):
                self.command.handle(**self.options)
        self.assertIn(INVALID_MODE_MSG.format(mode='2'), cm.output[0])

    def test_command_value_not_int(self):
        """Test blockprojectaccess with non-integer value (should fail)"""
        self.options['mode'] = 'abc'
        with self.assertLogs(self.logger_name, 'ERROR') as cm:
            with self.assertRaises(SystemExit):
                self.command.handle(**self.options)
        self.assertIn(
            MODE_CONVERT_ERR_MSG.format(mode='abc', ex=TypeError()),
            cm.output[0],
        )


class TestCleanAppSettings(
    ProjectMixin, RoleMixin, RoleAssignmentMixin, AppSettingMixin, TestCase
):
    """Tests for cleanappsettings command and associated functions"""

    def setUp(self):
        # Init roles
        self.init_roles()
        # Init users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.save()
        self.user_owner = self.make_user('owner')
        self.user_owner.email = 'owner_user@example.com'
        self.user_owner.save()

        # Init projects
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.plugin = Plugin.objects.get(name='example_project_app')

        # Init test settings
        self.setting_str_values = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_str_setting',
            'setting_type': APP_SETTING_TYPE_STRING,
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.setting_int_values = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_int_setting',
            'setting_type': APP_SETTING_TYPE_INTEGER,
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.setting_bool_values = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_bool_setting',
            'setting_type': APP_SETTING_TYPE_BOOLEAN,
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.setting_json_values = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_json_setting',
            'setting_type': APP_SETTING_TYPE_JSON,
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }
        self.setting_iprestrict_values = {
            'plugin_name': 'projectroles',
            'project': self.project,
            'name': 'ip_restrict',
            'setting_type': APP_SETTING_TYPE_BOOLEAN,
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.settings = [
            self.setting_int_values,
            self.setting_json_values,
            self.setting_str_values,
            self.setting_bool_values,
            self.setting_iprestrict_values,
        ]
        for s in self.settings:
            self.make_setting(
                plugin_name=s['plugin_name'],
                name=s['name'],
                setting_type=s['setting_type'],
                value=(
                    s['value']
                    if s['setting_type'] != APP_SETTING_TYPE_JSON
                    else ''
                ),
                value_json=(
                    s['value']
                    if s['setting_type'] == APP_SETTING_TYPE_JSON
                    else {}
                ),
                project=s['project'],
            )

        self.cmd_name = 'cleanappsettings'
        self.logger_name = LOGGER_PREFIX + self.cmd_name

    def test_command_undefined_project(self):
        """Test cleanappsettings with undefined PROJECT scope setting"""
        self.make_setting(
            plugin_name=self.plugin.name,
            name='undefined',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
        )
        self.assertEqual(AppSetting.objects.count(), 6)

        with self.assertLogs(self.logger_name, 'INFO') as cm:
            call_command(self.cmd_name)
            self.assertEqual(
                cm.output,
                [
                    CLEAN_LOG_PREFIX + START_MSG,
                    (
                        CLEAN_LOG_PREFIX
                        + DELETE_PREFIX_MSG.format(
                            s_name='settings.example_project_app.undefined',
                            project=f'"{self.project.title}"',
                            user=LOG_NONE_LABEL,
                        )
                        + DEF_NOT_FOUND_MSG
                    ),
                    CLEAN_LOG_PREFIX + END_MSG,
                ],
            )
        self.assertEqual(AppSetting.objects.count(), 5)
        self.assertIsNone(AppSetting.objects.filter(name='undefined').first())

    def test_command_undefined_user(self):
        """Test cleanappsettings with undefined USER scope setting"""
        self.make_setting(
            plugin_name=self.plugin.name,
            name='undefined',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            user=self.user_owner,
        )
        self.assertEqual(AppSetting.objects.count(), 6)
        call_command(self.cmd_name)
        self.assertEqual(AppSetting.objects.count(), 5)

    def test_command_undefined_project_user(self):
        """Test cleanappsettings with undefined PROJECT_USER scope setting"""
        self.make_setting(
            plugin_name=self.plugin.name,
            name='undefined',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
            user=self.user_owner,
        )
        self.assertEqual(AppSetting.objects.count(), 6)
        call_command(self.cmd_name)
        self.assertEqual(AppSetting.objects.count(), 5)

    def test_command_invalid_project_type(self):
        """Test cleanappsettings with invalid project type"""
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.category,
        )
        self.assertEqual(AppSetting.objects.count(), 6)

        with self.assertLogs(self.logger_name, 'INFO') as cm:
            call_command(self.cmd_name)
            self.assertEqual(len(cm.output), 3)
            self.assertEqual(
                cm.output[1],
                (
                    CLEAN_LOG_PREFIX
                    + DELETE_PREFIX_MSG.format(
                        s_name='settings.example_project_app.'
                        'project_user_bool_setting',
                        project=f'"{self.category.title}"',
                        user=LOG_NONE_LABEL,
                    )
                    + DELETE_PROJECT_TYPE_MSG.format(
                        self.category.type, ALLOWED_TYPES_MSG, 'PROJECT'
                    )
                ),
            )
        self.assertEqual(AppSetting.objects.count(), 5)
        self.assertIsNone(
            AppSetting.objects.filter(name='project_user_bool_setting').first()
        )

    def test_command_project_user_scope_no_role(self):
        """Test cleanappsettings with PROJECT_USER scope and no role"""
        user_new = self.make_user('user_new')
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
            user=user_new,
        )
        self.assertEqual(AppSetting.objects.count(), 6)

        with self.assertLogs(self.logger_name, 'INFO') as cm:
            call_command(self.cmd_name)
            self.assertEqual(len(cm.output), 3)
            self.assertEqual(
                cm.output[1],
                (
                    CLEAN_LOG_PREFIX
                    + DELETE_PREFIX_MSG.format(
                        s_name='settings.example_project_app.'
                        'project_user_bool_setting',
                        project=f'"{self.project.title}"',
                        user=f'"{user_new.username}"',
                    )
                    + DELETE_SCOPE_MSG.format(user_new.username)
                ),
            )
        self.assertEqual(AppSetting.objects.count(), 5)
        self.assertIsNone(
            AppSetting.objects.filter(name='project_user_bool_setting').first()
        )

    def test_command_project_user_scope_role(self):
        """Test cleanappsettings with PROJECT_USER scope and existing role"""
        user_new = self.make_user('user_new')
        self.make_assignment(self.project, user_new, self.role_contributor)
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
            user=user_new,
        )
        self.assertEqual(AppSetting.objects.count(), 6)
        call_command(self.cmd_name)
        self.assertEqual(AppSetting.objects.count(), 6)

    def test_command_project_user_superuser(self):
        """Test PROJECT_USER scope with no role as superuser"""
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
            user=self.superuser,
        )
        self.assertEqual(AppSetting.objects.count(), 6)

        with self.assertLogs(self.logger_name, 'INFO') as cm:
            call_command(self.cmd_name)
            self.assertEqual(len(cm.output), 2)
        self.assertEqual(AppSetting.objects.count(), 6)
        self.assertIsNotNone(
            AppSetting.objects.filter(name='project_user_bool_setting').first()
        )

    def test_command_project_user_superuser_enable(self):
        """Test PROJECT_USER scope with no role as superuser with deletion enabled"""
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
            user=self.superuser,
        )
        self.assertEqual(AppSetting.objects.count(), 6)

        with self.assertLogs(self.logger_name, 'INFO') as cm:
            call_command(self.cmd_name, superuser=True)
            self.assertEqual(len(cm.output), 3)
            self.assertEqual(
                cm.output[1],
                (
                    CLEAN_LOG_PREFIX
                    + DELETE_PREFIX_MSG.format(
                        s_name='settings.example_project_app.'
                        'project_user_bool_setting',
                        project=f'"{self.project.title}"',
                        user=f'"{self.superuser.username}"',
                    )
                    + DELETE_SCOPE_MSG.format(self.superuser.username)
                ),
            )
        self.assertEqual(AppSetting.objects.count(), 5)
        self.assertIsNone(
            AppSetting.objects.filter(name='project_user_bool_setting').first()
        )

    def test_command_check(self):
        """Test cleanappsettings with check mode enabled"""
        user_new = self.make_user('user_new')
        self.make_setting(
            plugin_name=self.plugin.name,
            name='undefined',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
        )
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.category,
        )
        self.make_setting(
            plugin_name=self.plugin.name,
            name='project_user_bool_setting',
            setting_type=APP_SETTING_TYPE_BOOLEAN,
            value=True,
            project=self.project,
            user=user_new,
        )
        self.assertEqual(AppSetting.objects.count(), 8)

        with self.assertLogs(self.logger_name, 'INFO') as cm:
            call_command(self.cmd_name, check=True)
            self.assertEqual(len(cm.output), 6)
        self.assertEqual(AppSetting.objects.count(), 8)


class TestRemoveRoles(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    RemoteSiteMixin,
    RemoteProjectMixin,
    TestCase,
):
    """Tests for removeroles command"""

    def setUp(self):
        # Init roles
        self.init_roles()
        # Init users
        self.user_owner_cat = self.make_user('user_owner_cat')
        self.user_owner = self.make_user('user_owner')
        self.user = self.make_user('user')
        # Init projects
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self.make_assignment(
            self.category, self.user_owner_cat, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        # Set other variables
        self.cmd_name = 'removeroles'

    def test_command(self):
        """Test removing non-owner roles from user"""
        # Set up non-owner roles for user in category and project
        self.make_assignment(self.category, self.user, self.role_contributor)
        self.make_assignment(self.project, self.user, self.role_guest)
        # Set up another user to ensure we keep their roles
        user_keep = self.make_user('user_keep')
        self.make_assignment(self.category, user_keep, self.role_contributor)
        self.make_assignment(self.project, user_keep, self.role_guest)

        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(self.category.has_role(user_keep), True)
        self.assertEqual(self.project.has_role(user_keep), True)
        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.assertEqual(self.user.is_active, True)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='role_delete').count(), 0
        )
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_delete').count(), 0
        )

        call_command(self.cmd_name, user=self.user)
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(self.category.has_role(user_keep), True)
        self.assertEqual(self.project.has_role(user_keep), True)
        self.assertEqual(self.category.has_role(self.user), False)
        self.assertEqual(self.project.has_role(self.user), False)
        self.user.refresh_from_db()
        self.assertEqual(self.user.is_active, True)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='role_delete').count(), 2
        )
        # App alerts should not be created
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_delete').count(), 0
        )

    def test_command_deactivate(self):
        """Test removing roles from user with deactivation"""
        # Set up non-owner roles for user in category and project
        self.make_assignment(self.category, self.user, self.role_contributor)
        self.make_assignment(self.project, self.user, self.role_guest)
        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.assertEqual(self.user.is_active, True)
        call_command(self.cmd_name, user=self.user, deactivate=True)
        self.assertEqual(self.category.has_role(self.user), False)
        self.assertEqual(self.project.has_role(self.user), False)
        self.user.refresh_from_db()
        self.assertEqual(self.user.is_active, False)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_command_remote(self):
        """Test removing roles from remote project"""
        self.make_assignment(self.category, self.user, self.role_contributor)
        self.make_assignment(self.project, self.user, self.role_guest)
        remote_site = self.make_site(
            name=REMOTE_SITE_NAME,
            url=REMOTE_SITE_URL,
            mode=SITE_MODE_SOURCE,
        )
        self.make_remote_project(
            project_uuid=self.category.sodar_uuid,
            site=remote_site,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.category,
        )
        self.make_remote_project(
            project_uuid=self.project.sodar_uuid,
            site=remote_site,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.project,
        )

        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.assertEqual(self.category.is_remote(), True)
        self.assertEqual(self.project.is_remote(), True)

        call_command(self.cmd_name, user=self.user)
        # Roles should not be removed
        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)

    def test_command_check(self):
        """Test command with check mode and non-owner role"""
        self.make_assignment(self.category, self.user, self.role_contributor)
        self.make_assignment(self.project, self.user, self.role_guest)

        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='role_delete').count(), 0
        )
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_delete').count(), 0
        )

        call_command(self.cmd_name, user=self.user, check=True)
        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.assertEqual(
            TimelineEvent.objects.filter(event_name='role_delete').count(), 0
        )
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_delete').count(), 0
        )

    def test_command_check_deactivate(self):
        """Test command with check mode and deactivation"""
        self.make_assignment(self.category, self.user, self.role_contributor)
        self.make_assignment(self.project, self.user, self.role_guest)
        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.assertEqual(self.user.is_active, True)
        call_command(self.cmd_name, user=self.user, check=True, deactivate=True)
        self.assertEqual(self.category.has_role(self.user), True)
        self.assertEqual(self.project.has_role(self.user), True)
        self.user.refresh_from_db()
        self.assertEqual(self.user.is_active, True)

    def test_command_owner_unset(self):
        """Test removing owner with unset new owner"""
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        # Inherited owner role
        self.assertEqual(self.project.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(
            TimelineEvent.objects.filter(
                event_name='role_owner_transfer'
            ).count(),
            0,
        )
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_update').count(), 0
        )

        call_command(self.cmd_name, user=self.user_owner)
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), False)
        # Assert local owner roles
        self.assertEqual(self.category.get_owner().user, self.user_owner_cat)
        self.assertEqual(self.project.get_owner().user, self.user_owner_cat)
        self.assertEqual(
            TimelineEvent.objects.filter(
                event_name='role_owner_transfer'
            ).count(),
            1,
        )
        # No alerts for inherited owner or old removed user
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_update').count(), 0
        )

    def test_command_owner_unset_top_level(self):
        """Test removing owner with unset new owner at top level (should fail)"""
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        call_command(self.cmd_name, user=self.user_owner_cat)
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)

    def test_command_owner_unset_same_owner(self):
        """Test removing owner with unset new owner and same owner on all levels (should fail)"""
        self.owner_as.user = self.user_owner_cat
        self.owner_as.save()
        self.assertEqual(self.category.get_owner().user, self.user_owner_cat)
        self.assertEqual(self.project.get_owner().user, self.user_owner_cat)
        call_command(self.cmd_name, user=self.user_owner_cat)
        self.assertEqual(self.category.get_owner().user, self.user_owner_cat)
        self.assertEqual(self.project.get_owner().user, self.user_owner_cat)

    def test_command_owner_set_no_role(self):
        """Test removing owner role with set new owner without existing role"""
        user_new = self.make_user('user_new')
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(self.project.has_role(user_new), False)
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_update').count(), 0
        )
        call_command(self.cmd_name, user=self.user_owner, owner=user_new)
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), False)
        self.assertEqual(self.project.get_owner().user, user_new)
        # Alert for new owner
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_update').count(), 1
        )

    def test_command_owner_set_no_role_top_level(self):
        """Test removing owner role with set new owner without existing role on top level"""
        user_new = self.make_user('user_new')
        self.assertEqual(self.category.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner_cat), True)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(self.project.has_role(user_new), False)
        call_command(self.cmd_name, user=self.user_owner_cat, owner=user_new)
        self.assertEqual(self.category.has_role(self.user_owner_cat), False)
        self.assertEqual(self.project.has_role(self.user_owner_cat), False)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(self.category.get_owner().user, user_new)
        self.assertEqual(self.project.has_role(user_new), True)

    def test_command_owner_set_existing_role(self):
        """Test removing owner role with set new owner and existing role"""
        user_new = self.make_user('user_new')
        self.make_assignment(self.project, user_new, self.role_contributor)
        self.assertEqual(self.project.get_owner().user, self.user_owner)
        self.assertEqual(self.project.has_role(user_new), True)
        self.assertEqual(
            RoleAssignment.objects.filter(user=user_new).count(), 1
        )
        call_command(self.cmd_name, user=self.user_owner, owner=user_new)
        self.assertEqual(self.project.get_owner().user, user_new)
        self.assertEqual(self.project.has_role(self.user_owner), False)
        self.assertEqual(
            RoleAssignment.objects.filter(user=user_new).count(), 1
        )

    def test_command_owner_invalid_new_owner_name(self):
        """Test removing owner with invalid new owner name (should fail)"""
        self.assertEqual(self.project.has_role(self.user_owner), True)
        with self.assertRaises(SystemExit):
            call_command(
                self.cmd_name, user=self.user_owner, owner='INVALID-NAME'
            )
        self.assertEqual(self.project.has_role(self.user_owner), True)

    def test_command_owner_check(self):
        """Test command with check mode and owner role"""
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(
            TimelineEvent.objects.filter(
                event_name='role_owner_transfer'
            ).count(),
            0,
        )
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_update').count(), 0
        )

        call_command(self.cmd_name, user=self.user_owner, check=True)
        self.assertEqual(self.project.has_role(self.user_owner), True)
        self.assertEqual(self.category.get_owner().user, self.user_owner_cat)
        self.assertEqual(self.project.get_owner().user, self.user_owner)
        self.assertEqual(
            TimelineEvent.objects.filter(
                event_name='role_owner_transfer'
            ).count(),
            0,
        )
        self.assertEqual(
            AppAlert.objects.filter(alert_name='role_update').count(), 0
        )


class TestSyncGroups(TestCase):
    """Tests for syncgroups command"""

    def setUp(self):
        self.command = SyncGroupsCommand()

    def test_sync_system(self):
        """Test sync with system username"""
        user = self.make_user('user')
        user.groups.all().delete()  # Remove groups
        self.assertEqual(user.groups.count(), 0)
        self.command.handle()
        self.assertEqual(user.groups.count(), 1)
        group = user.groups.first()
        self.assertIsNotNone(group)
        self.assertEqual(group.name, SYSTEM_USER_GROUP)

    def test_sync_system_existing(self):
        """Test sync with system username and existing group"""
        user = self.make_user('user')
        self.assertEqual(user.groups.count(), 1)
        self.command.handle()
        self.assertEqual(user.groups.count(), 1)
        group = user.groups.first()
        self.assertIsNotNone(group)
        self.assertEqual(group.name, SYSTEM_USER_GROUP)

    @override_settings(AUTH_LDAP_USERNAME_DOMAIN=LDAP_DOMAIN)
    def test_sync_ldap(self):
        """Test sync with LDAP user"""
        user = self.make_user('user@' + LDAP_DOMAIN)
        user.groups.all().delete()
        self.assertEqual(user.groups.count(), 0)
        self.command.handle()
        self.assertEqual(user.groups.count(), 1)
        group = user.groups.first()
        self.assertIsNotNone(group)
        self.assertEqual(group.name, LDAP_DOMAIN.lower())

    @override_settings(AUTH_LDAP_USERNAME_DOMAIN=LDAP_DOMAIN)
    def test_sync_ldap_existing(self):
        """Test sync with LDAP user and existing group"""
        user = self.make_user('user@' + LDAP_DOMAIN)
        self.assertEqual(user.groups.count(), 1)
        self.command.handle()
        self.assertEqual(user.groups.count(), 1)
        group = user.groups.first()
        self.assertIsNotNone(group)
        self.assertEqual(group.name, LDAP_DOMAIN.lower())


@override_settings(DEBUG=True)
class TestCreateDevUsers(TestCase):
    """Tests for createdevusers command"""

    def setUp(self):
        super().setUp()
        self.cmd_name = 'createdevusers'

    def test_create(self):
        """Test creating users"""
        self.assertEqual(User.objects.count(), 0)
        call_command(self.cmd_name)
        self.assertEqual(User.objects.count(), len(DEV_USER_NAMES))
        for u in DEV_USER_NAMES:
            user = User.objects.filter(
                username=u,
                name=f'{u.capitalize()} Example',
                first_name=u.capitalize(),
                last_name='Example',
                email=f'{u}@example.com',
            ).first()
            self.assertIsNotNone(user)
            self.assertEqual(user.check_password(DEFAULT_PASSWORD), True)

    def test_create_custom_password(self):
        """Test creating users with custom password"""
        call_command(self.cmd_name, password=CUSTOM_PASSWORD)
        for u in DEV_USER_NAMES:
            user = User.objects.filter(username=u).first()
            self.assertEqual(user.check_password(CUSTOM_PASSWORD), True)

    def test_create_existing(self):
        """Test creating users with existing user in list"""
        self.make_user(DEV_USER_NAMES[0])
        self.assertEqual(User.objects.count(), 1)
        call_command(self.cmd_name)
        self.assertEqual(User.objects.count(), len(DEV_USER_NAMES))

    def test_create_existing_not_in_list(self):
        """Test creating users with existing user not in list"""
        self.make_user('user_new')
        self.assertEqual(User.objects.count(), 1)
        call_command(self.cmd_name)
        self.assertEqual(User.objects.count(), len(DEV_USER_NAMES) + 1)

    @override_settings(DEBUG=False)
    def test_create_no_debug(self):
        """Test creating users with debug mode disabled (should fail)"""
        self.assertEqual(User.objects.count(), 0)
        with self.assertRaises(SystemExit):
            call_command(self.cmd_name)
        self.assertEqual(User.objects.count(), 0)
